#include "Matrix.h"

/*******************************************
 AllocMatrix :
   Allocation d'une matrice de dimension NxM
   avec initialisation a 0 des composantes
 USAGE :
   Matrix A;
   AllocMatrix(10,20,&A);
********************************************/
void AllocMatrix(unsigned int N, unsigned int M, Matrix *pA){
  double *pt;
  unsigned int i;
  CHECKALLOC( pA->val = (double **)malloc(N*sizeof(double*)) );
  PRINTMEMORYDEBUG(__func__,"double **",(void *)pA->val);
  CHECKALLOC( pt = (double *)calloc(N*M,sizeof(double)) );
  PRINTMEMORYDEBUG(__func__,"double *",(void *)pt);
  for (i=0;i<N;i++)
    pA->val[i] = pt + i*M;
  pA->contiguous=true;
  pA->rows=N;
  pA->cols=M;
}

/*******************************************
 AllocMatrix_out :
   Retourne une matrice de dimension NxM
   avec initialisation a 0 des composantes
 USAGE :
   Matrix A;
   A=AllocMatrix_out(10,20);
********************************************/
Matrix AllocMatrix_out(unsigned int N, unsigned int M){
  double *pt;
  unsigned int i;
  Matrix A;
  CHECKALLOC( A.val = (double **)malloc(N*sizeof(double*)) );
  PRINTMEMORYDEBUG(__func__,"double **",(void *)A.val);
  CHECKALLOC( pt = (double *)calloc(N*M,sizeof(double)) );
  PRINTMEMORYDEBUG(__func__,"double *",(void *)pt);
  for (i=0;i<N;i++)
    A.val[i] = pt + i*M;
  A.contiguous=true;
  A.rows=N;
  A.cols=M;
  return A;
}


/*******************************************
 secureAllocMatrix :
   Allocation (si besoin) d'une matrice de dimension NxM
   avec initialisation a 0 des composantes
 USAGE :
   Matrix A={0,0,NULL};
   ...
   secureAllocMatrix(10,20,&A);
********************************************/
void secureAllocMatrix(unsigned int N, unsigned int M, Matrix *pA){
  double *pt;
  unsigned int i;
  if ((pA->rows==N) && (pA->cols==M) && (pA->contiguous)){
      Matrix2zero(*pA);
      return;
  }
  if ((pA->rows>0) && (pA->cols>0))
    freeMatrix(*pA);
  CHECKALLOC( pA->val = (double **)malloc(N*sizeof(double*)) );
  PRINTMEMORYDEBUG(__func__,"double **",(void *)pA->val);
  CHECKALLOC( pt = (double *)calloc(N*M,sizeof(double)) );
  PRINTMEMORYDEBUG(__func__,"double *",(void *)pt);
  for (i=0;i<N;i++)
    pA->val[i] = pt + i*M;
  pA->contiguous=true;
  pA->rows=N;
  pA->cols=M;
}

/*******************************************
 AllocMatrixNC :
   Allocation d'une matrice de dimension NxM
   (Non Contigue) avec initialisation a 0 des composantes
 USAGE :
   Matrix A;
   AllocMatrixNC(10,20,&A);
********************************************/
void AllocMatrixNC(unsigned int N, unsigned int M, Matrix *pA){
  unsigned int i;
  CHECKALLOC( (*pA).val = (double **)malloc(N*sizeof(double*)) );
  for (i=0;i<N;i++)
    CHECKALLOC( (*pA).val[i] = (double *)calloc(M,sizeof(double)) );
  (*pA).contiguous=false;
  (*pA).rows=N;
  (*pA).cols=M;
}

/*******************************************
 AllocMatrixNC_out :
   Allocation d'une matrice de dimension NxM
   (Non Contigue) avec initialisation a 0 des composantes
 USAGE :
   Matrix A;
   A=AllocMatrixNC_out(10,20);
********************************************/
Matrix AllocMatrixNC_out(unsigned int N, unsigned int M){
  unsigned int i;
  Matrix A;
  CHECKALLOC( A.val = (double **)malloc(N*sizeof(double*)) );
  for (i=0;i<N;i++)
    CHECKALLOC( A.val[i] = (double *)calloc(M,sizeof(double)) );
  A.contiguous=false;
  A.rows=N;
  A.cols=M;
  return A;
}

/*******************************************
 secureAllocMatrixNC :
   Allocation (si besoin) d'une matrice de dimension NxM
   (Non Contigue) avec initialisation a 0 des composantes
 USAGE :
   Matrix A={0,0,NULL};
   secureAllocMatrixNC(10,20,&A);
 REMARQUE ;
   Pas si secure...
********************************************/
void secureAllocMatrixNC(unsigned int N, unsigned int M, Matrix *pA){
  unsigned int i;
  if ((pA->rows==N) && (pA->cols==M) && (!pA->contiguous)){
      Matrix2zero(*pA);
      return;
  }
  if ((pA->rows>0) && (pA->cols>0))
    freeMatrix(*pA);
  CHECKALLOC( (*pA).val = (double **)malloc(N*sizeof(double*)) );
  for (i=0;i<N;i++)
    CHECKALLOC( (*pA).val[i] = (double *)calloc(M,sizeof(double)) );
  (*pA).contiguous=false;
  (*pA).rows=N;
  (*pA).cols=M;
}

/*******************************************
 freeMatrix :
   Desalloue simplement une matrice :
 USAGE :
   Matrix A;
   ...
   freeMatrix(A);
 REMARQUE:
   Utiliser de preference FREEMatrix
********************************************/
void freeMatrix(Matrix Mat){
  int i;
  if (Mat.contiguous){
    double *pt=Mat.val[0];
    PRINTMEMORYDEBUG(__func__,"double *",(void *)pt);
    free(pt);
  }
  else
    for (i=0;i<Mat.rows;i++){
      PRINTMEMORYDEBUG(__func__,"double *",(void *)Mat.val[i]);
      free(Mat.val[i]);
    }
  PRINTMEMORYDEBUG(__func__,"double **",(void *)Mat.val);
  free(Mat.val);
}

/*******************************************
 FREEMatrix :
   Desalloue proprement une matrice :
   <val> <- NULL
   <nL>  <- 0
   <nC>  <- 0
 USAGE :
   Matrix A;
   ...
   FREEMatrix(&A);
********************************************/
void FREEMatrix(Matrix *pMat){
  int i;
  if ((*pMat).contiguous)
    free((*pMat).val[0]);
  else
    for (i=0;i<(*pMat).rows;i++)
      free((*pMat).val[i]);
  free((*pMat).val);
  (*pMat).val=NULL;
  (*pMat).rows=0;
  (*pMat).cols=0;
}

/*******************************************
 RandMatrix :
   Initialise les composantes d'une matrice
   suivant la loi uniforme sur [a;b].
   La matrice est suppose deja alloue.
 USAGE :
   Matrix A;
   ...
   RandMatrix(A,-1.,1.);
********************************************/
void RandMatrix(Matrix M, double a, double b){
  unsigned int i;
  if (M.contiguous)
    uniform_real_distribution(a,b,M.rows*M.cols,M.val[0]);
  else
    for (i=0;i<M.rows;i++)
      uniform_real_distribution(a,b,M.rows*M.cols,M.val[i]);
}

/*******************************************
 AllocRandMatrix :
   Alloue une matrice de dimension NxM et
   initialise les composantes d'une matrice
   suivant la loi uniforme sur [a;b].
   La matrice est suppose non-allouee.
 USAGE :
   Matrix A;
   AllocRandMatrix(10,20,-1.,1.,&A);
********************************************/
void AllocRandMatrix(unsigned int N, unsigned int M, double a, double b,Matrix *pA){
  AllocMatrix(N,M,pA);
  RandMatrix(*pA,a,b);
}

/*******************************************
 AllocRandMatrix_out :
   Alloue une matrice de dimension NxM et
   initialise les composantes d'une matrice
   suivant la loi uniforme sur [a;b].
   La matrice est suppose non-allouee.
 USAGE :
   Matrix A;
   A=AllocRandMatrix_out(10,20,-1.,1.);
********************************************/
Matrix AllocRandMatrix_out(unsigned int N, unsigned int M, double a, double b){
  Matrix A;
  A=AllocMatrix_out(N,M);
  RandMatrix(A,a,b);
  return A;
}

/*******************************************
 MatrixProd_out :
   Calcul la Matrice Q <- A*B
   ou  A,B sont deux matrices.
   Le matrice Q est allouee (si besoin) dans cette fonction
 USAGE :
   Matrix U,V,W;
   ...
   W=MatrixProd_out(U,V);
 REMARQUE :
   Que se passe-t'il avec
   U=MatrixProd_out(U,V);
********************************************/
Matrix MatrixProd_out(Matrix A, Matrix B){
  unsigned int i,j,k;
  double s;
  Matrix Q;
  MESSAGEFAILURE( A.cols != B.rows, "dimensions failed\n");
  Q=AllocMatrix_out(A.rows,B.cols);
  for (i=0;i<Q.rows;i++)
    for (j=0;j<Q.cols;j++){
      s=0;
      for (k=0;k<A.cols;k++)
        s+=A.val[i][k]*B.val[k][j];
      Q.val[i][j]=s;
  }
  return Q;
}

/*******************************************
 MatrixProd :
   Calcul la Matrice Q <- A*B
   ou  A,B sont deux matrices.
   Le matrice Q est allouee (si besoin) dans cette fonction
 USAGE :
   Matrix U,V,W;
   ...
   MatrixProd(U,V,&W);
 REMARQUE :
   Que se passe-t'il avec
   MatrixProd(U,V,&U);
********************************************/
void MatrixProd(Matrix A, Matrix B, Matrix *pQ){
  unsigned int i,j,k;
  double s;
  MESSAGEFAILURE( A.cols != B.rows, "dimensions failed\n");
  AllocMatrix(A.rows,B.cols,pQ);
  for (i=0;i<pQ->rows;i++)
    for (j=0;j<pQ->cols;j++){
      s=0;
      for (k=0;k<A.cols;k++)
        s+=A.val[i][k]*B.val[k][j];
      pQ->val[i][j]=s;
  }
}

/*******************************************
 MatrixSum :
   Calcul la Matrice Q <- a*A+b*B
   ou a,b sont 2 reels et A,B deux matrices.
   Le matrice Q est déjà allouee. (Q peut-etre
   la matrice A ou la matrice B)
 USAGE :
   Matrix U,V,W;
   ...
   MatrixSum(-1,U,1.,V,Q);
 REMARQUE :
   Que se passe-t'il avec
     MatrixSum(-1,U,1,V,U);
   ou
     MatrixSum(-1,U,1,V,V);
   ou
     MatrixSum(-1,V,1,V,V);
********************************************/
void MatrixSum(double a, Matrix A,double b, Matrix B, Matrix Q){
  unsigned int i,j;
  MESSAGEFAILURE( (A.cols != B.cols) || (A.rows != B.rows) , "dimensions failed\n");
  MESSAGEFAILURE( (A.cols != Q.cols) || (A.rows != Q.rows) , "dimensions failed\n");
  for (i=0;i<Q.rows;i++)
    for (j=0;j<Q.cols;j++)
      Q.val[i][j]=a*A.val[i][j]+b*B.val[i][j];
}

/*******************************************
 MatrixSumAlloc :
   Calcul la Matrice Q <- a*A+b*B
   ou a,b sont 2 reels et A,B deux matrices.
   La matrice Q est allouee (si besoin) dans cette fonction
 USAGE :
   Matrix U,V,W;
   ...
   MatrixSumAlloc(-1,U,1.,V,&W);
 REMARQUE :
   Que se passe-t'il avec
     MatrixSumAlloc(-1,U,1,V,&U);
   ou
     MatrixSumAlloc(-1,U,1,V,&V);
   ou
     MatrixSumAlloc(-1,V,1,V,&V);
********************************************/
void MatrixSumAlloc(double a, Matrix A,double b, Matrix B, Matrix *pQ){
  unsigned int i,j;
  MESSAGEFAILURE( (A.cols != B.cols) || (A.rows != B.rows) , "dimensions failed\n");
  AllocMatrix(A.rows,A.cols,pQ);
  for (i=0;i<A.rows;i++)
    for (j=0;j<A.cols;j++)
      pQ->val[i][j]=a*A.val[i][j]+b*B.val[i][j];
}

/*******************************************
 MatrixSum_out :
   Calcul la Matrice Q <- a*A+b*B
   ou a,b sont 2 reels et A,B deux matrices.
   La matrice Q est allouee (si besoin) dans cette fonction
 USAGE :
   Matrix U,V,W;
   ...
   W=MatrixSum_out(-1,U,1.,V);
 REMARQUE :
   Que se passe-t'il avec
     U=MatrixSum_out(-1,U,1,V);
   ou
     V=MatrixSum_out(-1,U,1,V);
   ou
     V=MatrixSum_out(-1,V,1,V);
********************************************/
Matrix MatrixSum_out(double a, Matrix A,double b, Matrix B){
  unsigned int i,j;
  Matrix Q;
  MESSAGEFAILURE( (A.cols != B.cols) || (A.rows != B.rows) , "dimensions failed\n");
  AllocMatrix(A.rows,A.cols,&Q);
  for (i=0;i<A.rows;i++)
    for (j=0;j<A.cols;j++)
      Q.val[i][j]=a*A.val[i][j]+b*B.val[i][j];
  return Q;
}

/*******************************************
 MatrixSumLocLeft :
   Transforme la matrice A en  a*A+b*B :
     A <- a*A+b*B
   ou a,b sont 2 reels et A,B deux matrices.
 USAGE :
   Matrix U,V;
   ...
   MatrixSumLocLeft(-1,U,1.,V);
********************************************/
void MatrixSumLocLeft(double a, Matrix A,double b, Matrix B){
  unsigned int i,j;
  MESSAGEFAILURE( (A.cols != B.cols) || (A.rows != B.rows) , "dimensions failed\n");
  for (i=0;i<A.rows;i++)
    for (j=0;j<A.cols;j++)
      A.val[i][j]=a*A.val[i][j]+b*B.val[i][j];
}

/*******************************************
 MatrixSumLocRight :
   Transforme la matrice B en  a*A+b*B :
     B <- a*A+b*B
   ou a,b sont 2 reels et A,B deux matrices.
 USAGE :
   Matrix U,V;
   ...
   MatrixSumLocLeft(-1,U,1.,V);
********************************************/
void MatrixSumLocRight(double a, Matrix A,double b, Matrix B){
  unsigned int i,j;
  MESSAGEFAILURE( (A.cols != B.cols) || (A.rows != B.rows) , "dimensions failed\n");
  for (i=0;i<A.rows;i++)
    for (j=0;j<A.cols;j++)
      B.val[i][j]=a*A.val[i][j]+b*B.val[i][j];
}


void PrintLine(Matrix A,unsigned int i,unsigned int nc,const char *format){
  unsigned int j;
  if (2*nc>=A.cols)
    for (j=0;j<A.cols;j++)
      printf(format,A.val[i][j]);
    else{
      for (j=0;j<nc;j++)
        printf(format,A.val[i][j]);
      printf("...  ");
      for (j=A.cols-nc;j<A.cols;j++)
        printf(format,A.val[i][j]);
    }
  printf("\n");
}

void PrintLineName(Matrix A,char *VarName,unsigned int i,unsigned int nc,const char *format){
  printf("  %s[%d]<"ANSI_COLOR_BLUE"%p"ANSI_COLOR_RESET">: ",VarName,i,(void *)A.val[i]);
  //printf("  %s[%d]<\033[0;34m%p\033[0m>: ",VarName,i,(void *)A.val[i]);
  PrintLine(A,i,nc,format);
}
/*******************************************
 PrintMatrix :
   Affiche partiellement une matrice A
     A(1:nl,1:nc) i.e. nr premieres lignes avec nc premieres colonnes
     puis
     A(end-nl:end,1:nc) i.e. nr dernieres lignes avec nc premieres colonnes
 USAGE :
   Matrix A;
   ...
   PrintMatrix(A,"A",2,5);
********************************************/
void PrintMatrix(Matrix A,char *VarName,unsigned int nr,unsigned int nc){
  PrintMatrixFormat(A,VarName,nr,nc,"%.6f");
}

void PrintMatrixFormat(Matrix A,char *VarName,unsigned int nr,unsigned int nc,const char *format){
  unsigned int i;
  char sformat[1024];
  strcpy(sformat,format);
  strcat(sformat," ");
  if (nr==0) nr=A.rows;
  if (nc==0) nc=A.cols;
  printf("%s<"ANSI_COLOR_BLUE"%p"ANSI_COLOR_RESET">[%d][%d] :\n",VarName,(void *)A.val,A.rows,A.cols);
  //printf("%s<\033[0;34m%p\033[0m>[%d][%d] :\n",VarName,(void *)A.val,A.rows,A.cols);
  if (2*nr>=A.rows)
    for (i=0;i<A.rows;i++)
      PrintLineName(A,VarName,i,nc,sformat);
  else{
    for (i=0;i<nr;i++)
      PrintLineName(A,VarName,i,nc,sformat);
    printf("  ...\n");
    for (i=A.rows-nr;i<A.rows;i++)
      PrintLineName(A,VarName,i,nc,sformat);
  }
}

void PrintMatrix1(Matrix A,char *VarName){
  unsigned int i,j;
  printf("Matrice %s (%dx%d):\n",VarName,A.rows,A.cols);
  for (i=0;i<A.rows;i++){
    printf(" %d: ",i);
    for (j=0;j<A.cols;j++)
      if (A.val[i][j]<0)
        printf("%lf ",A.val[i][j]);
      else
  printf(" %lf ",A.val[i][j]);
    printf("\n");
  }
}

/*******************************************
 IdentityMatrix :
   Genere la matrice identite
 USAGE :
   Matrix I;
   IdentityMatrix(20,&I);
********************************************/
void IdentityMatrix(unsigned int N, Matrix *pI){
  unsigned int i;
  AllocMatrix(N,N,pI);
  for (i=0;i<pI->rows;i++)
    pI->val[i][i]=1.;
}

/*******************************************
 IdentityMatrix_out :
   Genere la matrice identite
 USAGE :
   Matrix I;
   I=IdentityMatrix_out(20);
********************************************/
Matrix IdentityMatrix_out(unsigned int N){
  unsigned int i;
  Matrix I;
  AllocMatrix(N,N,&I);
  for (i=0;i<I.rows;i++)
    I.val[i][i]=1.;
  return I;
}

/*******************************************
 Matrix2cst:
   Met la matrice a c
 USAGE :
   Matrix A;
   ...
   Matrix2cst(A,1.);
********************************************/
void Matrix2cst(Matrix A, double c){
  unsigned int i,j;
  for (i=0;i<A.rows;i++)
    for (j=0;j<A.cols;j++)
      A.val[i][j]=c;
}

/*******************************************
 Matrix2zero:
   Met la matrice a 0
 USAGE :
   Matrix A
   ...;
   Matrix2zero(A);
********************************************/
void Matrix2zero(Matrix A){
  Matrix2cst(A,0.);
}



/*******************************************
 ScaleMatrixLoc :
   Transforme la matrice A en a*A :
      A <- a*A
   ou a est un scalaire
 USAGE :
   Matrix A;
   ...
   ScaleMatrix(2.,A);
********************************************/
void ScaleMatrix(double a,Matrix A){
  unsigned int i,j;

  for (i=0;i<A.rows;i++)
    for (j=0;j<A.cols;j++)
      A.val[i][j]*=a;
}

/*******************************************
 TransposeMatrix :
   Genere la matrice transposee d'une matrice
 USAGE :
   Matrix A,At;
   ...
   TransposeMatrix(A,&At);
********************************************/
void TransposeMatrix(Matrix A, Matrix *pAt){
  unsigned int i,j;
  AllocMatrix(A.cols,A.rows,pAt);
  for (i=0;i<pAt->rows;i++)
    for (j=0;j<pAt->cols;j++)
      pAt->val[i][j]=A.val[j][i];
}

/*******************************************
 TransposeMatrix_out :
   Retourne la matrice transposee d'une matrice
 USAGE :
   Matrix A,At;
   ...
   At=TransposeMatrix_out(A);
********************************************/
Matrix TransposeMatrix_out(Matrix A){
  unsigned int i,j;
  Matrix At;
  AllocMatrix(A.cols,A.rows,&At);
  for (i=0;i<At.rows;i++)
    for (j=0;j<At.cols;j++)
      At.val[i][j]=A.val[j][i];
  return At;
}

/*******************************************
 AllocRandSPDMatrix :
   Genere une matrice aleatoire symetrique definie positive
 USAGE :
   Matrix A;
   AllocRandSPDMatrix(10,&A);
********************************************/
void AllocRandSPDMatrix(unsigned int N, Matrix *pB){
  Matrix At,A;
  AllocRandTriLowerMatrix(N,1.,3.,&A);
  TransposeMatrix(A,&At);
  MatrixProd(A,At,pB);
  freeMatrix(A);
  freeMatrix(At);
}

/*******************************************
 AllocRandSPDMatrix_out :
   Retourne une matrice aleatoire symetrique definie positive
 USAGE :
   Matrix A;
   A=AllocRandSPDMatrix_out(10);
********************************************/
Matrix AllocRandSPDMatrix_out(unsigned int N){
  Matrix At,A,B;
  AllocRandTriLowerMatrix(N,1.,3.,&A);
  TransposeMatrix(A,&At);
  MatrixProd(A,At,&B);
  freeMatrix(A);
  freeMatrix(At);
  return B;
}

/*******************************************
 AllocRandTriUpperMatrix :
   Genere une matrice aleatoire triangulaire superieure
   suivant la loi uniforme sur (a,b)
 USAGE :
   Matrix A;
   AllocRandTriUpperMatrix(10,-2,2,&A);
********************************************/
void AllocRandTriUpperMatrix(unsigned int N,double a, double b, Matrix *pQ){
  unsigned int i,j,R,k=0;
  double *pR;
  AllocMatrix(N,N,pQ);
  R=N*(N+1)/2;
  CHECKALLOC( pR = (double *)calloc(R,sizeof(double)) );
  uniform_real_distribution(a,b,R,pR);
  for (i=0;i<N;i++)
    for (j=i;j<N;j++)
      pQ->val[i][j]=pR[k++];
  free(pR);
}

/*******************************************
 AllocRandTriUpperMatrix_out :
   Retourne une matrice carree aleatoire triangulaire superieure
   suivant la loi uniforme sur (a,b)
 USAGE :
   Matrix A;
   AllocRandTriUpperMatrix(10,-2,2,&A);
********************************************/
Matrix AllocRandTriUpperMatrix_out(unsigned int N,double a, double b){
  unsigned int i,j,R,k=0;
  double *pR;
  Matrix Q;
  AllocMatrix(N,N,&Q);
  R=N*(N+1)/2;
  CHECKALLOC( pR = (double *)calloc(R,sizeof(double)) );
  uniform_real_distribution(a,b,R,pR);
  for (i=0;i<N;i++)
    for (j=i;j<N;j++)
      Q.val[i][j]=pR[k++];
  free(pR);
  return Q;
}

/*******************************************
 AllocRandTriUpperMatrix :
   Genere une matrice aleatoire triangulaire inferieure
   suivant la loi uniforme sur (a,b)
 USAGE :
   Matrix A;
   AllocRandTriUpperMatrix(10,-2,2,&A);
********************************************/
void AllocRandTriLowerMatrix(unsigned int N,double a, double b, Matrix *pQ){
  unsigned int i,j,R,k=0;
  double *pR;
  AllocMatrix(N,N,pQ);
  R=N*(N+1)/2;
  CHECKALLOC( pR = (double *)calloc(R,sizeof(double)) );
  uniform_real_distribution(a,b,R,pR);
  for (i=0;i<N;i++)
    for (j=0;j<=i;j++)
      pQ->val[i][j]=pR[k++];
  free(pR);
}

/*******************************************
 AllocRandTriUpperMatrix_out :
   Genere une matrice aleatoire triangulaire inferieure
   suivant la loi uniforme sur (a,b)
 USAGE :
   Matrix A;
   A=AllocRandTriUpperMatrix_out(10,-2,2);
********************************************/
Matrix AllocRandTriLowerMatrix_out(unsigned int N,double a, double b){
  unsigned int i,j,R,k=0;
  double *pR;
  Matrix Q;
  AllocMatrix(N,N,&Q);
  R=N*(N+1)/2;
  CHECKALLOC( pR = (double *)calloc(R,sizeof(double)) );
  uniform_real_distribution(a,b,R,pR);
  for (i=0;i<N;i++)
    for (j=0;j<=i;j++)
      Q.val[i][j]=pR[k++];
  free(pR);
  return Q;
}

/*******************************************
 Norm1Matrix :
   Calcule ||A||_1 (norme 1 matricielle)
********************************************/
double Norm1Matrix(Matrix A){
  unsigned int i,j;
  double max=0.,s;

  for (j=0;j<A.cols;j++){
    s=0.;
    for (i=0;i<A.rows;i++)
      s+=fabs(A.val[i][j]);
    max=MAX(s,max);
  }
  return max;
}

/*******************************************
 NormInfMatrix :
   Calcule ||A||_\infty (norme infini matricielle)
********************************************/
double NormInfMatrix(Matrix A){
  unsigned int i,j;
  double max=0.,s;

  for (i=0;i<A.rows;i++){
    s=0.;
    for (j=0;j<A.cols;j++)
      s+=fabs(A.val[i][j]);
    max=MAX(s,max);
  }
  return max;
}

/*******************************************
 fevalMatrix :
   Alloue (si besoin) la matrice B tel que
       B[i][j] <- f(A[i][j])
 USAGE :
   Matrix A,B;

   ...
   fevalVector(A,exp,&B);
********************************************/
void fevalMatrix(Matrix A,double (*pf)(double), Matrix *pB){
  unsigned int i,j;
  AllocMatrix(A.rows,A.cols,pB);
  for (i=0;i<A.rows;i++)
    for (j=0;j<A.cols;j++)
      pB->val[i][j]=pf(A.val[i][j]);
}

/*******************************************
 fevalMatrix_out :
   Retourne la matrice B tel que
       B[i][j] <- f(A[i][j])
 USAGE :
   Matrix A,B;

   ...
   B=fevalVector_out(A,exp);
********************************************/
Matrix fevalMatrix_out(Matrix A,double (*pf)(double)){
  unsigned int i,j;
  Matrix B;
  AllocMatrix(A.rows,A.cols,&B);
  for (i=0;i<A.rows;i++)
    for (j=0;j<A.cols;j++)
      B.val[i][j]=pf(A.val[i][j]);
  return B;
}

/*******************************************
 fevalMatrix_in :
   Modifie la matrice A
       A[i][j] <- f(A[i][j])
 USAGE :
   Matrix A;

   ...
   fevalVector_in(A,exp);
********************************************/
void fevalMatrix_in(Matrix A,double (*pf)(double)){
  unsigned int i,j;
  for (i=0;i<A.rows;i++)
    for (j=0;j<A.cols;j++)
      A.val[i][j]=pf(A.val[i][j]);
}


/*******************************************
 copyMatrix :
   Copie intégrale d'une matrice avec duplication
   des tableaux.
 USAGE :
   Matrix A,B;

   ...
   copyMatrix(A,&B);
********************************************/
void copyMatrix(Matrix A, Matrix *pB){
  unsigned int i;
  AllocMatrix(A.rows,A.cols,pB);
  if (A.contiguous)
    memcpy(pB->val[0],A.val[0],A.rows*A.cols*sizeof(double));
  else{
    for (i=0;i<A.rows;i++)
      memcpy(pB->val[i],A.val[i],A.cols*sizeof(double));
  }
}

/*******************************************
 copyMatrix_out :
   Copie intégrale d'une matrice avec duplication
   des tableaux.
 USAGE :
   Matrix A,B;

   ...
   copyMatrix(A,&B);
********************************************/
Matrix copyMatrix_out(Matrix A){
  unsigned int i;
  Matrix B;
  AllocMatrix(A.rows,A.cols,&B);
  if (A.contiguous)
    memcpy(B.val[0],A.val[0],A.rows*A.cols*sizeof(double));
  else{
    for (i=0;i<A.rows;i++)
      memcpy(B.val[i],A.val[i],A.cols*sizeof(double));
  }
  return B;
}
