#ifndef _MATRIX_H_
#define _MATRIX_H_

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h> // since C99
#include <math.h>
#include <string.h>
#include "slacgen.h"
#include "rand.h"

typedef struct {
 unsigned int rows;
 unsigned int cols;
 double **val;
 bool contiguous;
 } Matrix;
 
#define M0 {0,0,NULL,0}

/*******************************************
 AllocMatrix :
   Allocation d'une matrice de dimension NxM 
   avec initialisation a 0 des composantes 
 USAGE :                                 
   Matrix A;
   AllocMatrix(10,20,&A);                      
********************************************/
void AllocMatrix(unsigned int N, unsigned int M, Matrix *pA);

/*******************************************
 AllocMatrix_out :
   Retourne une matrice de dimension NxM 
   avec initialisation a 0 des composantes 
 USAGE :                                 
   Matrix A;
   A=AllocMatrix_out(10,20);                      
********************************************/
Matrix AllocMatrix_out(unsigned int N, unsigned int M);


/*******************************************
 secureAllocMatrix :
   Allocation (si besoin) d'une matrice de dimension NxM 
   avec initialisation a 0 des composantes 
 USAGE :                                 
   Matrix A={0,0,NULL};
   ...
   secureAllocMatrix(10,20,&A);                      
********************************************/
void secureAllocMatrix(unsigned int N, unsigned int M, Matrix *pA);

/*******************************************
 AllocMatrixNC :
   Allocation d'une matrice de dimension NxM 
   (Non Contigue) avec initialisation a 0 des composantes 
 USAGE :                                 
   Matrix A;
   AllocMatrixNC(10,20,&A);                      
********************************************/
void AllocMatrixNC(unsigned int N, unsigned int M, Matrix *pA);

/*******************************************
 AllocMatrixNC_out :
   Allocation d'une matrice de dimension NxM 
   (Non Contigue) avec initialisation a 0 des composantes 
 USAGE :                                 
   Matrix A;
   A=AllocMatrixNC_out(10,20);                      
********************************************/
Matrix AllocMatrixNC_out(unsigned int N, unsigned int M);

/*******************************************
 secureAllocMatrixNC :
   Allocation (si besoin) d'une matrice de dimension NxM 
   (Non Contigue) avec initialisation a 0 des composantes  
 USAGE :                                 
   Matrix A={0,0,NULL};
   secureAllocMatrixNC(10,20,&A);
 REMARQUE ;
   Pas si secure...
********************************************/
void secureAllocMatrixNC(unsigned int N, unsigned int M, Matrix *pA);

/*******************************************
 freeMatrix :
   Desalloue simplement une matrice : 
 USAGE :                                 
   Matrix A;
   ...
   freeMatrix(A);
 REMARQUE:
   Utiliser de preference FREEMatrix
********************************************/
void freeMatrix(Matrix Mat);
 
/*******************************************
 FREEMatrix :
   Desalloue proprement une matrice : 
   <val> <- NULL
   <nL>  <- 0
   <nC>  <- 0
 USAGE :                                 
   Matrix A;
   ...
   FREEMatrix(&A);                      
********************************************/
void FREEMatrix(Matrix *pMat);

/*******************************************
 RandMatrix :
   Initialise les composantes d'une matrice
   suivant la loi uniforme sur [a;b].
   La matrice est suppose deja alloue.
 USAGE :                                 
   Matrix A;
   ...
   RandMatrix(A,-1.,1.);                      
********************************************/
void RandMatrix(Matrix M, double a, double b);

/*******************************************
 AllocRandMatrix :
   Alloue une matrice de dimension NxM et
   initialise les composantes d'une matrice
   suivant la loi uniforme sur [a;b].
   La matrice est suppose non-allouee.
 USAGE :                                 
   Matrix A;
   AllocRandMatrix(10,20,-1.,1.,&A);                      
********************************************/
void AllocRandMatrix(unsigned int N, unsigned int M, double a, double b,Matrix *pA);

/*******************************************
 AllocRandMatrix_out :
   Alloue une matrice de dimension NxM et
   initialise les composantes d'une matrice
   suivant la loi uniforme sur [a;b].
   La matrice est suppose non-allouee.
 USAGE :                                 
   Matrix A;
   A=AllocRandMatrix_out(10,20,-1.,1.);                      
********************************************/
Matrix AllocRandMatrix_out(unsigned int N, unsigned int M, double a, double b);

/*******************************************
 MatrixProd_out :
   Calcul la Matrice Q <- A*B
   ou  A,B sont deux matrices.
   Le matrice Q est allouee (si besoin) dans cette fonction
 USAGE :                                 
   Matrix U,V,W;
   ...
   W=MatrixProd_out(U,V);
 REMARQUE :
   Que se passe-t'il avec
   U=MatrixProd_out(U,V);
********************************************/
Matrix MatrixProd_out(Matrix A, Matrix B);

/*******************************************
 MatrixProd :
   Calcul la Matrice Q <- A*B
   ou  A,B sont deux matrices.
   Le matrice Q est allouee (si besoin) dans cette fonction
 USAGE :                                 
   Matrix U,V,W;
   ...
   MatrixProd(U,V,&W);
 REMARQUE :
   Que se passe-t'il avec
   MatrixProd(U,V,&U);
********************************************/
void MatrixProd(Matrix A, Matrix B, Matrix *pQ);

/*******************************************
 MatrixSum :
   Calcul la Matrice Q <- a*A+b*B
   ou a,b sont 2 reels et A,B deux matrices.
   Le matrice Q est déjà allouee. (Q peut-etre
   la matrice A ou la matrice B)
 USAGE :                                 
   Matrix U,V,W;
   ...
   MatrixSum(-1,U,1.,V,Q);                      
 REMARQUE :
   Que se passe-t'il avec
     MatrixSum(-1,U,1,V,U);   
   ou  
     MatrixSum(-1,U,1,V,V);
   ou
     MatrixSum(-1,V,1,V,V);
********************************************/
void MatrixSum(double a, Matrix A,double b, Matrix B, Matrix Q);

/*******************************************
 MatrixSumAlloc :
   Calcul la Matrice Q <- a*A+b*B
   ou a,b sont 2 reels et A,B deux matrices.
   La matrice Q est allouee (si besoin) dans cette fonction
 USAGE :                                 
   Matrix U,V,W;
   ...
   MatrixSumAlloc(-1,U,1.,V,&W);
 REMARQUE :
   Que se passe-t'il avec
     MatrixSumAlloc(-1,U,1,V,&U);   
   ou  
     MatrixSumAlloc(-1,U,1,V,&V);
   ou
     MatrixSumAlloc(-1,V,1,V,&V);
********************************************/
void MatrixSumAlloc(double a, Matrix A,double b, Matrix B, Matrix *pQ);

/*******************************************
 MatrixSum_out :
   Calcul la Matrice Q <- a*A+b*B
   ou a,b sont 2 reels et A,B deux matrices.
   La matrice Q est allouee (si besoin) dans cette fonction
 USAGE :                                 
   Matrix U,V,W;
   ...
   W=MatrixSum_out(-1,U,1.,V);
 REMARQUE :
   Que se passe-t'il avec
     U=MatrixSum_out(-1,U,1,V);   
   ou  
     V=MatrixSum_out(-1,U,1,V);
   ou
     V=MatrixSum_out(-1,V,1,V);
********************************************/
Matrix MatrixSum_out(double a, Matrix A,double b, Matrix B);

/*******************************************
 MatrixSumLocLeft :
   Transforme la matrice A en  a*A+b*B :
     A <- a*A+b*B 
   ou a,b sont 2 reels et A,B deux matrices.
 USAGE :                                 
   Matrix U,V;
   ...
   MatrixSumLocLeft(-1,U,1.,V);                      
********************************************/
void MatrixSumLocLeft(double a, Matrix A,double b, Matrix B);

/*******************************************
 MatrixSumLocRight :
   Transforme la matrice B en  a*A+b*B :
     B <- a*A+b*B 
   ou a,b sont 2 reels et A,B deux matrices.
 USAGE :                                 
   Matrix U,V;
   ...
   MatrixSumLocLeft(-1,U,1.,V);                      
********************************************/
void MatrixSumLocRight(double a, Matrix A,double b, Matrix B);


/*******************************************
 PrintMatrix :
   Affiche partiellement une matrice A
     A(1:nl,1:nc) i.e. nr premieres lignes avec nc premieres colonnes
     puis
     A(end-nl:end,1:nc) i.e. nr dernieres lignes avec nc premieres colonnes
 USAGE :                                 
   Matrix A;
   ...
   PrintMatrix(A,"A",2,5);                      
********************************************/
void PrintMatrix(Matrix A,char *VarName,unsigned int nr,unsigned int nc);

void PrintMatrixFormat(Matrix A,char *VarName,unsigned int nr,unsigned int nc,const char *format);


/*******************************************
 IdentityMatrix :
   Genere la matrice identite
 USAGE :                                 
   Matrix I;
   IdentityMatrix(20,&I);
********************************************/
void IdentityMatrix(unsigned int N, Matrix *pI);

/*******************************************
 IdentityMatrix_out :
   Genere la matrice identite
 USAGE :                                 
   Matrix I;
   I=IdentityMatrix_out(20);
********************************************/
Matrix IdentityMatrix_out(unsigned int N);

/*******************************************
 Matrix2cst:
   Met la matrice a c
 USAGE :                                 
   Matrix A;
   ...
   Matrix2cst(A,1.);
********************************************/
void Matrix2cst(Matrix A, double c);

/*******************************************
 Matrix2zero:
   Met la matrice a 0
 USAGE :                                 
   Matrix A
   ...;
   Matrix2zero(A);
********************************************/
void Matrix2zero(Matrix A);

/*******************************************
 ScaleMatrixLoc :
   Transforme la matrice A en a*A :
      A <- a*A
   ou a est un scalaire
 USAGE :                                 
   Matrix A;
   ...
   ScaleMatrix(2.,A);
********************************************/
void ScaleMatrix(double a,Matrix A);

/*******************************************
 TransposeMatrix :
   Genere la matrice transposee d'une matrice
 USAGE :                                 
   Matrix A,At;
   ...
   TransposeMatrix(A,&At);
********************************************/
void TransposeMatrix(Matrix A, Matrix *pAt);

/*******************************************
 TransposeMatrix_out :
   Retourne la matrice transposee d'une matrice
 USAGE :                                 
   Matrix A,At;
   ...
   At=TransposeMatrix_out(A);
********************************************/
Matrix TransposeMatrix_out(Matrix A);

/*******************************************
 AllocRandSPDMatrix :
   Genere une matrice aleatoire symetrique definie positive
 USAGE :                                 
   Matrix A;
   AllocRandSPDMatrix(10,&A);
********************************************/
void AllocRandSPDMatrix(unsigned int N, Matrix *pB);

/*******************************************
 AllocRandSPDMatrix_out :
   Retourne une matrice aleatoire symetrique definie positive
 USAGE :                                 
   Matrix A;
   A=AllocRandSPDMatrix_out(10);
********************************************/
Matrix AllocRandSPDMatrix_out(unsigned int N);

/*******************************************
 AllocRandTriUpperMatrix :
   Genere une matrice aleatoire triangulaire superieure 
   suivant la loi uniforme sur (a,b)
 USAGE :                                 
   Matrix A;
   AllocRandTriUpperMatrix(10,-2,2,&A);
********************************************/
void AllocRandTriUpperMatrix(unsigned int N,double a, double b, Matrix *pQ);

/*******************************************
 AllocRandTriUpperMatrix_out :
   Retourne une matrice aleatoire triangulaire superieure 
   suivant la loi uniforme sur (a,b)
 USAGE :                                 
   Matrix A;
   AllocRandTriUpperMatrix(10,-2,2,&A);
********************************************/
Matrix AllocRandTriUpperMatrix_out(unsigned int N,double a, double b);

/*******************************************
 AllocRandTriUpperMatrix :
   Genere une matrice aleatoire triangulaire inferieure 
   suivant la loi uniforme sur (a,b)
 USAGE :                                 
   Matrix A;
   AllocRandTriUpperMatrix(10,-2,2,&A);
********************************************/
void AllocRandTriLowerMatrix(unsigned int N,double a, double b, Matrix *pQ);

/*******************************************
 AllocRandTriUpperMatrix_out :
   Genere une matrice aleatoire triangulaire inferieure 
   suivant la loi uniforme sur (a,b)
 USAGE :                                 
   Matrix A;
   A=AllocRandTriUpperMatrix_out(10,-2,2);
********************************************/
Matrix AllocRandTriLowerMatrix_out(unsigned int N,double a, double b);

/*******************************************
 Norm1Matrix :
   Calcule ||A||_1 (norme 1 matricielle) 
********************************************/
double Norm1Matrix(Matrix A);

/*******************************************
 NormInfMatrix :
   Calcule ||A||_\infty (norme infini matricielle) 
********************************************/
double NormInfMatrix(Matrix A);

/*******************************************
 fevalMatrix :
   Alloue (si besoin) la matrice B tel que  
       B[i][j] <- f(A[i][j])
 USAGE :                                 
   Matrix A,B;
   
   ...
   fevalVector(A,exp,&B);
********************************************/
void fevalMatrix(Matrix A,double (*pf)(double), Matrix *pB);

/*******************************************
 fevalMatrix_out :
   Retourne la matrice B tel que  
       B[i][j] <- f(A[i][j])
 USAGE :                                 
   Matrix A,B;
   
   ...
   B=fevalVector_out(A,exp);
********************************************/
Matrix fevalMatrix_out(Matrix A,double (*pf)(double));

/*******************************************
 fevalMatrix_in :
   Modifie la matrice A  
       A[i][j] <- f(A[i][j])
 USAGE :                                 
   Matrix A;
   
   ...
   fevalVector_in(A,exp);
********************************************/
void fevalMatrix_in(Matrix A,double (*pf)(double));


/*******************************************
 copyMatrix :
   Copie intégrale d'une matrice avec duplication
   des tableaux.
 USAGE :                                 
   Matrix A,B;
   
   ...
   copyMatrix(A,&B);
********************************************/
void copyMatrix(Matrix A, Matrix *pB);

/*******************************************
 copyMatrix_out :
   Copie intégrale d'une matrice avec duplication
   des tableaux.
 USAGE :                                 
   Matrix A,B;
   
   ...
   copyMatrix(A,&B);
********************************************/
Matrix copyMatrix_out(Matrix A);


#endif
