
/*********************************************************
                     S.L.A.C
           Simple Linear Algebra in C

  Loin d'etre optimale, ces fonctions permettent
  d'illustrer les notions de base pouvant etre vues dans
  un cours de "Langage C pour le calcul scientifique".
  De nombreuses fonctions sont redondantes et sont
  proposees a titre pedagogique.

                               (c) cuvelier 2019
**********************************************************/
#include "SLAC.h"

/*******************************************
 MatrixProd :
   Calcul le vecteur  V=A*U
   ou  A est une matrice et U un vecteur
   Le vecteur V est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector U,V;
   ...
   MatVecProd(A,U,&V);
********************************************/
void MatVecProd(Matrix A, Vector U, Vector *pV){
  if(A.cols != U.dim ){//|| pV->dim != A.rows
    fprintf(stderr, "Erreur dans la dimention de les dimentions d'entrées dans %s\n",__func__ );
    exit(EXIT_FAILURE);

  }
  AllocVector(A.rows,pV);
  for(int i = 0; i < A.rows; i++){
    for(int j = 0; j < A.cols; j++){
      pV->val[i]+=A.val[i][j]*U.val[j];
    }
  }
}

/*******************************************
 MatrixProd_out :
   Retourne le vecteur A*U
   ou  A est une matrice et U un vecteur
   Le vecteur resultat est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector U,V;
   ...
   V=MatVecProd_out(A,U);
********************************************/
Vector MatVecProd_out(Matrix A, Vector U){
  if(A.cols != U.dim){
    fprintf(stderr, "Erreur dans la dimention de la matrice et du vecteur dans %s\n",__func__ );
    exit(EXIT_FAILURE);
  }
  Vector w = AllocVector_out(A.rows);
  for(int i = 0; i < A.rows; i++){
    for(int j = 0; j < A.cols; j++){
      w.val[i] += A.val[i][j]*U.val[j];
    }
  }
  return w;
}


/*******************************************
 SolveLower :
   Resoud le systeme lineaire L*x=b
   ou  L est une matrice triangulaire inferieure
   et x,b deux vecteurs
   Le vecteur x est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector b,x;
   ...
   SolveLower((A,b,&x);
********************************************/
void SolveLower(Matrix L, Vector b, Vector *px){
  if (L.rows!=L.cols ) {
    fprintf(stderr, "Error in matrix or vector size in %s\n",__func__ );
    exit(EXIT_FAILURE);
  }
  AllocVector(b.dim, px);
  int n = L.rows;
  for(int i = 0; i <= n-1; i++){
    double S = 0;
    for(int j = 0; j <= i-1 ; j++){
      S+=L.val[i][j]*px->val[j];
    }
    px->val[i] = (b.val[i]-S)/L.val[i][i];
  }
}


/*******************************************
 SolveLower_out :
   Retourne la solution x du systeme lineaire L*x=b
   ou  L est une matrice triangulaire inferieure
   et x,b deux vecteurs
   Le vecteur x est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector b,x;
   ...
   x=SolveLower((A,b);
********************************************/
Vector SolveLower_out(Matrix L, Vector b){
  if(L.rows != L.cols || L.rows != b.dim){
    fprintf(stderr, "Error in matrix or vector size in %s\n",__func__ );
  }
  Vector  V =  AllocVector_out(L.rows);

  int n = L.rows;
  for(int i = 0; i <= n-1; i++){
    double S = 0;
    for(int j = 0; j <= i-1 ; j++){
      S+=L.val[i][j]*V.val[j];
    }
    V.val[i] = (b.val[i]-S)/L.val[i][i];
  }
  return V;
}

/*******************************************
 SolveUpper :
   Resoud le systeme lineaire U*x=b
   ou  U est une matrice triangulaire superieure
   et x,b deux vecteurs
   Le vecteur x est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector b,x;
   ...
   SolveUpper(A,b,&x);
********************************************/
void SolveUpper(Matrix U, Vector b, Vector *px){
  if (U.rows!=U.cols) {
    fprintf(stderr, "Error in matrix or vector size in %s\n",__func__ );
    exit(EXIT_FAILURE);
  }
  AllocVector(b.dim,px);
  int n = U.rows;
  for(int i = n-1; i >=0; i--){
    double S = 0;
    for(int j = i; j <= n-1; j++){
      S += U.val[i][j]*px->val[j];
    }
    px->val[i] = (b.val[i]-S)/U.val[i][i];
  }
}

/*******************************************
 SolveUpper_out :
   Retourne la solution du systeme lineaire U*x=b
   ou  U est une matrice triangulaire superieure
   et x,b deux vecteurs
   Le vecteur x est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector b,x;
   ...
   x=SolveUpper_out(A,b);
********************************************/
Vector SolveUpper_out(Matrix U, Vector b){
  if(U.rows != U.cols || U.rows != b.dim){
    fprintf(stderr, "Error in matrix or vector size in %s\n",__func__ );
  }
  Vector  V =  AllocVector_out(U.rows);
  int n = U.rows;
  for(int i = n-1; i >-1; i--){
    double S = 0;
    for(int j = i; j <= n-1; j++){
      S += U.val[i][j]*V.val[j];
    }
    V.val[i] = (b.val[i]-S)/U.val[i][i];
  }
  return V;
}

/*******************************************
 Cholesky :
   Calcul la matrice L telle que A=L*L^t
   ou  A est une matrice symetrique definie
   positive et L une matrice triangulaire
   L est la factorisation positive de Cholesky.
   inferieure a diagonale positive.
   Le matrice L est allouee dans cette fonction
 USAGE :
   Matrix A,L;
   ...
   Cholesky(A,&L);
********************************************/
void Cholesky(Matrix A, Matrix *L){

  AllocMatrix(A.rows,A.cols,L);
  int N = A.rows ;
  int i = 0 ; int j = 0 ; int k = 0 ;
  double S1 = 0.0 ; double S2 = 0.0 ;
  for(i=0;i<N;i=i+1){
    S1 = 0.0 ;
    for(j=0;j<i;j=j+1){
      S1 = S1 + fabs( (*L).val[i][j] ) * fabs( (*L).val[i][j] ) ;
    }
    (*L).val[i][i] = sqrt(A.val[i][i] - S1) ;

    for(j=0;j<i;j=j+1){
      (*L).val[j][i] = 0.0 ;
    }
    for(j=i+1;j<N;j=j+1){
      S2 = 0.0 ;
      for(k=0;k<i;k=k+1){
        S2 = S2 + fabs( (*L).val[j][k] ) * fabs( (*L).val[i][k] ) ;
      }
      (*L).val[j][i] = (A.val[j][i] - S2) / (*L).val[i][i] ;
    }
  }
}
/*******************************************
 Cholesky_out :
   Retourne la matrice L telle que A=L*L^t
   ou  A est une matrice symetrique definie
   positive et L une matrice triangulaire
   inferieure a diagonale positive.
   L est la factorisation positive de Cholesky.
   Le matrice L est allouee dans cette fonction
 USAGE :
   Matrix A,L;
   ...
   L=Cholesky(A);
********************************************/
Matrix Cholesky_out(Matrix A)
{
  int N = A.rows ;
  Matrix L = AllocMatrix_out(N,N);
  Cholesky(A,&L);
  return(L);
}
/*******************************************
 SolveCholesky :
   Resoud le systeme A*x=b
   ou A=L*L^t est une matrice symetrique
   definie positive et L une matrice
   triangulaire inferieure a diagonale
   positive. L est la factorisation
   positive de Cholesky.
   x et b sont deux vecteurs
 USAGE :
   Matrix A,L;
   Vector x,b;
   ...
   Cholesky(A,&L);
   SolveCholesky(L,b,&x);
********************************************/
void SolveCholesky(Matrix L, Vector b, Vector *x)
{
  if( L.cols != b.dim || L.rows != (*x).dim || b.dim != (*x).dim )
  {
    fprintf(stderr, "Error in matrix or vector size in %s\n",__func__ );
  }

  int N = b.dim ;

  Vector y =  AllocVector_out(N);

  y = SolveLower_out(L,b);

  Matrix L_transp = AllocMatrix_out(N,N);
  L_transp = TransposeMatrix_out(L);

  (*x) = SolveUpper_out(L_transp,y);

  FREEMatrix(&L_transp);
  FREEVector(&y);
}
/*******************************************
 SolveCholesky_out :
   Retourne la solution du systeme A*x=b
   ou A=L*L^t est une matrice symetrique
   definie positive et L une matrice
   triangulaire inferieure a diagonale
   positive. L est la factorisation
   positive de Cholesky.
   x et b sont deux vecteurs
 USAGE :
   Matrix A,L;
   Vector x,b;
   ...
   L=Cholesky_out(A);
   x=SolveCholesky_out(L,b);
********************************************/
Vector SolveCholesky_out(Matrix L, Vector b){
  Vector x = AllocVector_out(b.dim);
  SolveCholesky(L,b,&x);
  return(x);
}
/*******************************************
 LU :
   Calcul les matrices L et U telle que A=L*U
   ou  A est une matrice,
   L une matrice triangulaire inferieure a
   diagonale unite et U une matrice triangulaire
   superieure. L et U sont la factorisation
   LU de la matrice A.
   Les matrices L et U sont allouees dans cette
   fonction
 USAGE :
   Matrix A,L,U;
   ...
   LU(A,&L,&U);
********************************************/
void LU(Matrix A, Matrix *pL, Matrix *pU)
{
  if(A.rows != A.cols){
    fprintf(stderr, "Erreur in dimentions of matrix function %s\n", __func__);
  }
  int n = A.rows;
  /* on s'assure que la matrice A est carrée*/
  AllocMatrix( n, n, pL);
  AllocMatrix( n, n, pU);
  IdentityMatrix(A.rows, pL); //initialisation de L a la matrice identite
  Matrix2zero(*pU); // initialisation de U a la matrice nulle

  double somme;
  int i,j,k;
  /* Vérifier que A hermitienne definie positive => chacune de ses sous-matrices principales sont inversibles*/
  for(i = 0; i < n ; i++)
  {
    for(j = i; j < n ; j++)
    {
      somme = 0;
      for(k = 0; k < i; k++)
      {
        somme = somme + pL->val[i][k] * pU->val[k][j];
      }
      pU->val[i][j] = A.val[i][j]-somme;
    }
    for (j = i+1; j < n ; j++)
    {
      somme = 0;
      for (k = 0; k < i; k++)
      {
        somme = somme +  pL->val[j][k] * pU->val[k][i];
      }
      pL->val[j][i] = (A.val[j][i] - somme)/pU->val[i][i];
    }
  }
}

/*******************************************
 SolveLU :
   Resoud le systeme A*x=b
   ou A=L*U est une matrice dont les sous-matrices
   principales sont inversibles.
   L une matrice triangulaire inferieure a
   diagonale unite et U une matrice triangulaire
   superieure.
   x et b sont deux vecteurs
 USAGE :
   Matrix A,L,U;
   Vector x,b;
   ...
   LU(A,&L,&U);
   SolveLU(L,U,b,&x);
********************************************/
void SolveLU(Matrix L, Matrix U, Vector b, Vector *px)
{
  Vector V = AllocVector_out(L.rows);
  V = SolveLower_out( L, b);
  SolveUpper(U, V, px);
}

/*******************************************
 SolveLU_out :
   Retourne la solution du systeme A*x=b
   ou A=L*U est une matrice dont les sous-matrices
   principales sont inversibles.
   L une matrice triangulaire inferieure a
   diagonale unite et U une matrice triangulaire
   superieure.
   x et b sont deux vecteurs
 USAGE :
   Matrix A,L,U;
   Vector x,b;
   ...
   LU(A,&L,&U);
   x=SolveLU_out(L,U,b);
********************************************/
Vector SolveLU_out(Matrix L, Matrix U, Vector b)
{
  Vector px = AllocVector_out(L.rows);
  Vector V = AllocVector_out(L.rows);
  V = SolveLower_out( L, b);
  px = SolveUpper_out(U , V);
  return px;
}

/*******************************************
ChercheIndPivot :
  Cherche un pivot pour la matrice A
  afin d'effectuer le pivot de Gauss

USAGE :
  Matric A;
  int j =... , pivot;
  ...
  pivot = ChercheIntPivot(A,j);
********************************************/

int ChercheIndPivot(Matrix A, int j)
{
  int k=j;
  int n= A.rows;
  double pivot=fabs(A.val[j-1][j-1]);
  for (int i=j+1;j<=n;j++)
  {
    if (fabs(A.val[i-1][j-1])>pivot)
    {
      k=i;
      pivot=fabs(A.val[i-1][j-1]);
    }
  }
  return k;
}



/*******************************************
PermLigneSys :
  Pivote la ligne j et k
********************************************/


void PermLigneSys(Matrix A, Vector b, int j, int k)
{
  int l;
  double t;
   int n= A.rows;
  for (l=1;l<=n;l++)
  {
    t=A.val[j-1][l-1];
    A.val[j-1][l-1]=A.val[k-1][l-1];
    A.val[k-1][l-1]=t;
  }
  t=b.val[j-1];
  b.val[j-1]=b.val[k-1];
  b.val[k-1]=t;
}
/*******************************************
********************************************/


void CombLigneSys (Matrix A, Vector b, int j, int i, double alpha)
{
    int n= A.rows;
    for (int k=1;k<=n;k++)
    {
        A.val[i-1][k-1]=A.val[i-1][k-1]+alpha*A.val[j-1][k-1];
    }
    b.val[i-1]=b.val[i-1]+alpha*b.val[j-1];
}

/*******************************************
SolveGauss_in :
Resoud le systeme A*x=b
par la methode de Gauss avec pivot partiel
x et b sont deux vecteurs.
Attention la matrice A et le vecteur b
sont modifies.
USAGE :
Matrix A;
Vector x,b;
...
SolveGauss_in(A,b,&x);
********************************************/
void SolveGauss_in(Matrix A, Vector b){
    int k;
    int n=A.rows;
      for(int j=1;j<=n-1;j++){
        k=ChercheIndPivot(A,j);
        PermLigneSys(A,b,j,k);
        for(int i=j+1;i<=n;i++){
            CombLigneSys(A,b,j,i,-A.val[i-1][j-1]/(A.val[j-1][j-1]));
        }
    }
    SolveUpper(A,b,&b);

}

/*******************************************
 SolveGauss :
   Resoud le systeme A*x=b
   par la methode de Gauss avec pivot partiel
   x et b sont deux vecteurs.
   Attention la matrice A et le vecteur b
   sont modifies.
 USAGE :
   Matrix A;
   Vector x,b;
   ...
   SolveGauss(A,b,&x);
********************************************/
void SolveGauss(Matrix A, Vector b,Vector* px){
    int k;
    int n=A.rows;
    for(int j=1;j<=n-1;j++){
        k=ChercheIndPivot(A,j);
        PermLigneSys(A,b,j,k);
        for(int i=j+1;i<=n;i++){
            CombLigneSys(A,b,j,i,-A.val[i-1][j-1]/(A.val[j-1][j-1]));
        }
    }
    SolveUpper(A,b,px);

 }
/*******************************************
 SolveGauss_out :
   Retourne la solution du systeme A*x=b
   par la methode de Gauss avec pivot partiel
   x et b sont deux vecteurs.
   La matrice A et le vecteur b ne doivent pas
   etre modifies.
 USAGE :
   Matrix A;
   Vector x,b;
   ...
   x=SolveGauss_out(A,b);
********************************************/
Vector SolveGauss_out(Matrix A, Vector b){
    Vector x;
    AllocVector(A.rows,&x);
    Matrix D;
    AllocMatrix(A.rows,A.cols,&D);
    D=A;
    Vector e;
    AllocVector(A.rows,&e);
    e=b;
    int k;
    int n=A.rows;
    for(int j=1;j<=n-1;j++){
        k=ChercheIndPivot(D,j);
        PermLigneSys(D,e,j,k);
        for(int i=j+1;i<=n;i++){
            CombLigneSys(D,e,j,i,-D.val[i-1][j-1]/(D.val[j-1][j-1]));
        }
    }
    SolveUpper(D,e,&x);

    return x;
}
