#ifndef _SLAS_H_
#define _SLAS_H_

/*********************************************************
                     S.L.A.C
           Simple Linear Algebra in C

  Loin d'etre optimale, ces fonctions permettent
  d'illustrer les notions de base pouvant etre vues dans
  un cours de "Langage C pour le calcul scientifique".
  De nombreuses fonctions sont redondantes et sont
  proposees a titre pedagogique.

                               (c) cuvelier 2019
**********************************************************/
#include "Vector.h"
#include "Matrix.h"

/*******************************************
 MatrixProd :
   Calcul le vecteur  V=A*U
   ou  A est une matrice et U un vecteur
   Le vecteur V est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector U,V;
   ...
   MatVecProd(A,U,&V);
********************************************/
void MatVecProd(Matrix A, Vector U, Vector *pV);

/*******************************************
 MatrixProd_out :
   Retourne le vecteur A*U
   ou  A est une matrice et U un vecteur
   Le vecteur resultat est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector U,V;
   ...
   V=MatVecProd_out(A,U);
********************************************/
Vector MatVecProd_out(Matrix A, Vector U);


/*******************************************
 SolveLower :
   Resoud le systeme lineaire L*x=b
   ou  L est une matrice triangulaire inferieure
   et x,b deux vecteurs
   Le vecteur x est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector b,x;
   ...
   SolveLower((A,b,&x);
********************************************/
void SolveLower(Matrix L, Vector b, Vector *px);

/*******************************************
 SolveLower_out :
   Retourne la solution x du systeme lineaire L*x=b
   ou  L est une matrice triangulaire inferieure
   et x,b deux vecteurs
   Le vecteur x est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector b,x;
   ...
   x=SolveLower((A,b);
********************************************/
Vector SolveLower_out(Matrix L, Vector b);

/*******************************************
 SolveUpper :
   Resoud le systeme lineaire U*x=b
   ou  U est une matrice triangulaire superieure
   et x,b deux vecteurs
   Le vecteur x est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector b,x;
   ...
   SolveUpper(A,b,&x);
********************************************/
void SolveUpper(Matrix U, Vector b, Vector *px);

/*******************************************
 SolveUpper_out :
   Retourne la solution du systeme lineaire U*x=b
   ou  U est une matrice triangulaire superieure
   et x,b deux vecteurs
   Le vecteur x est allouee dans cette fonction
 USAGE :
   Matrix A;
   Vector b,x;
   ...
   x=SolveUpper_out(A,b);
********************************************/
Vector SolveUpper_out(Matrix U, Vector b);

/*******************************************
 Cholesky :
   Calcul la matrice L telle que A=L*L^t
   ou  A est une matrice symetrique definie
   positive et L une matrice triangulaire
   inferieure a diagonale positive.
   L est la factorisation positive de Cholesky.
   Le matrice L est allouee dans cette fonction
 USAGE :
   Matrix A,L;
   ...
   Cholesky(A,&L);
********************************************/
void Cholesky(Matrix A, Matrix *pL);

/*******************************************
 Cholesky_out :
   Retourne la matrice L telle que A=L*L^t
   ou  A est une matrice symetrique definie
   positive et L une matrice triangulaire
   inferieure a diagonale positive.
   L est la factorisation positive de Cholesky.
   Le matrice L est allouee dans cette fonction
 USAGE :
   Matrix A,L;
   ...
   L=Cholesky(A);
********************************************/
Matrix Cholesky_out(Matrix A);

/*******************************************
 SolveCholesky :
   Resoud le systeme A*x=b
   ou A=L*L^t est une matrice symetrique
   definie positive et L une matrice
   triangulaire inferieure a diagonale
   positive. L est la factorisation
   positive de Cholesky.
   x et b sont deux vecteurs
 USAGE :
   Matrix A,L;
   Vector x,b;
   ...
   Cholesky(A,&L);
   SolveCholesky(L,b,&x);
********************************************/
void SolveCholesky(Matrix L, Vector b, Vector *px);

/*******************************************
 SolveCholesky_out :
   Retourne la solution du systeme A*x=b
   ou A=L*L^t est une matrice symetrique
   definie positive et L une matrice
   triangulaire inferieure a diagonale
   positive. L est la factorisation
   positive de Cholesky.
   x et b sont deux vecteurs
 USAGE :
   Matrix A,L;
   Vector x,b;
   ...
   L=Cholesky_out(A);
   x=SolveCholesky_out(L,b);
********************************************/
Vector SolveCholesky_out(Matrix L, Vector b);

/*******************************************
 LU :
   Calcul les matrices L et U telle que A=L*U
   ou  A est une matrice,
   L une matrice triangulaire inferieure a
   diagonale unite et U une matrice triangulaire
   superieure. L et U sont la factorisation
   LU de la matrice A.
   Les matrices L et U sont allouees dans cette
   fonction
 USAGE :
   Matrix A,L,U;
   ...
   LU(A,&L,&U);
********************************************/
void LU(Matrix A, Matrix *pL, Matrix *pU);

/*******************************************
 SolveLU :
   Resoud le systeme A*x=b
   ou A=L*U est une matrice dont les sous-matrices
   principales sont inversibles.
   L une matrice triangulaire inferieure a
   diagonale unite et U une matrice triangulaire
   superieure.
   x et b sont deux vecteurs
 USAGE :
   Matrix A,L,U;
   Vector x,b;
   ...
   LU(A,&L,&U);
   SolveLU(L,U,b,&x);
********************************************/
void SolveLU(Matrix L, Matrix U, Vector b, Vector *px);

/*******************************************
 SolveLU_out :
   Retourne la solution du systeme A*x=b
   ou A=L*U est une matrice dont les sous-matrices
   principales sont inversibles.
   L une matrice triangulaire inferieure a
   diagonale unite et U une matrice triangulaire
   superieure.
   x et b sont deux vecteurs
 USAGE :
   Matrix A,L,U;
   Vector x,b;
   ...
   LU(A,&L,&U);
   x=SolveLU_out(L,U,b);
********************************************/
Vector SolveLU_out(Matrix L, Matrix U, Vector b);

/*******************************************
 SolveGauss_in :
   Resoud le systeme A*x=b
   par la methode de Gauss avec pivot partiel
   x et b sont deux vecteurs.
   Attention la matrice A et le vecteur b
   sont modifies.
 USAGE :
   Matrix A;
   Vector x,b;
   ...
   SolveGauss_in(A,b,&x);
********************************************/
int ChercheIndPivot(Matrix A, int j);
void PermLigneSys(Matrix A, Vector b, int j, int k);
void CombLigneSys (Matrix A, Vector b, int j, int i, double alpha);

void SolveGauss_in(Matrix A, Vector b);

/*******************************************
 SolveGauss :
   Resoud le systeme A*x=b
   par la methode de Gauss avec pivot partiel
   x et b sont deux vecteurs.
   Attention la matrice A et le vecteur b
   sont modifies.
 USAGE :
   Matrix A;
   Vector x,b;
   ...
   SolveGauss(A,b,&x);
********************************************/
void SolveGauss(Matrix A, Vector b, Vector *px);

/*******************************************
 SolveGauss_out :
   Retourne la solution du systeme A*x=b
   par la methode de Gauss avec pivot partiel
   x et b sont deux vecteurs.
   La matrice A et le vecteur b ne doivent pas
   etre modifies.
 USAGE :
   Matrix A;
   Vector x,b;
   ...
   x=SolveGauss_out(A,b);
********************************************/
Vector SolveGauss_out(Matrix A, Vector b);








#endif
