#include <math.h>
#include <assert.h>

#include "Vector.h"

/*******************************************
 AllocVector_out :
   Retourne un vecteur de dimension N
   avec initialisation a 0 des composantes
 USAGE :
   Vector V;
   ...
   V=AllocVector_out(10);
  REMARQUE :
    Que se passe-t'il pour:
    Vector V;
    ...
    V=AllocVector_out(10);
    V=AllocVector_out(20);

    fuite mémoire:  utiliser valgrind
********************************************/
Vector AllocVector_out(unsigned int N){
  Vector U;
  CHECKALLOC( U.val = (double *)calloc(N,sizeof(double)) );
  PRINTMEMORYDEBUG(__func__,"double *",(void *)U.val);
  U.dim=N;
  return U;
}


/*******************************************
 AllocVector :
   Allocation d'un vecteur de dimension N
   avec initialisation a 0 des composantes
 USAGE :
   Vector V;
   AllocVector(10,&V);

 REMARQUE :
    Que se passe-t'il pour:
    Vector V;
    ...
    AllocVector1(10,&V);
    AllocVector1(20,&V);

    fuite mémoire:  utiliser valgrind
********************************************/
void AllocVector(unsigned int N, Vector *pU){
  CHECKALLOC( pU->val = (double *)calloc(N,sizeof(double)) );
  PRINTMEMORYDEBUG(__func__,"double *",(void *)pU->val);
  pU->dim=N;
}

/*******************************************
 secureAllocVector :
   Allocation d'un vecteur de dimension N
   avec initialisation a 0 des composantes.
   Si le vecteur a deja ete alloue et est
   de dimension N alors juste une simple mise
   à 0 du vecteur.
 USAGE :
   Vector V={0,NULL};
   ...
   secureAllocVector(10,&V);
  REMARQUE :
    n'est pas si secure que cela!
      Vector V;
      secureAllocVector(10,&V);

    => Core dump!
    Il faut initialiser tous les vecteurs a 0
    pour que cette fonction soit operationnelle:
      Vector U={0,NULL},V={0,NULL};
********************************************/
void secureAllocVector(unsigned int N, Vector *pU){
  if (pU->dim!=N){
    if (pU->dim>0 )
      freeVector(*pU);
    CHECKALLOC( pU->val = (double *)calloc(N,sizeof(double)) );
    PRINTMEMORYDEBUG(__func__,"double *",(void *)pU->val);
    pU->dim=N;
  }else{
    Vector2zero(*pU);
  }
}

/*******************************************
 FREEVector :
   Desalloue proprement un vecteur :
   <val> <- NULL
   <dim>  <- 0
 USAGE :
   Vector V;
   FREEVector(&V);
********************************************/
void FREEVector(Vector *pV){
  if (pV->dim>0){
    PRINTMEMORYDEBUG(__func__,"double *",(void *)pV->val);
    free(pV->val);
    pV->val=NULL;
    pV->dim=0;
  }else{
    MESSAGEWARNING( "Cannot free an empty Vector");
  }
}

/*******************************************
 FREEVector :
   Desalloue simplement un vecteur
 USAGE :
   Vector V;
   freeVector(V);
********************************************/
void freeVector(Vector V){
  PRINTMEMORYDEBUG(__func__,"double *",(void *)V.val);
  free(V.val);
}

/*******************************************
 Vector2zero :
   Met les composantes d'un vecteur à 0
 USAGE :
   Vector V;
   ...
   Vector2zero(V);
********************************************/
void Vector2zero(Vector V){
  int i;
  for (i=0;i<V.dim;i++) V.val[i]=0;
}

/*******************************************
 Vector2cst :
   Met les composantes d'un vecteur à c
 USAGE :
   Vector V;
   ...
   Vector2cst(V,1.);
********************************************/
void Vector2cst(Vector V,double c){
  int i;
  for (i=0;i<V.dim;i++) V.val[i]=c;
}


/*******************************************
 AllocRandVector :
   Alloue (si besoin) et initialise les composantes d'un vecteur
   suivant la loi uniforme sur [a;b].
 USAGE :
   Vector V;
   ...
   AllocRandVector(N,-1.,1.,&V);
********************************************/
void AllocRandVector(int N, double a, double b,Vector *pW){
  AllocVector(N,pW);
  uniform_real_distribution(a,b,N,pW->val);
}


/*******************************************
 RandVector :
   Initialise les composantes d'un vecteur
   suivant la loi uniforme sur [a;b].
   Le vecteur est suppose deja alloue.
 USAGE :
   Vector V;
   ...
   RandVector(V,-1.,1.);
********************************************/
void RandVector(Vector U, double a, double b){
  uniform_real_distribution(a,b,U.dim,U.val);
}

/*******************************************
 AllocRandVector_out :
   Alloue un vecteur de dimension N et
   initialise les composantes d'un vecteur
   suivant la loi uniforme sur [a;b].
 USAGE :
   Vector V;
   V=AllocRandVector_out(10,-1.,1.);
 REMARQUE :
   Attention aux fuites mémoire
********************************************/
Vector AllocRandVector_out(unsigned int N, double a, double b){
  Vector U;
  AllocVector(N,&U);
  uniform_real_distribution(a,b,U.dim,U.val);
  return U;
}

/*******************************************
 PrintVector :
   Affiche un vecteur
 USAGE :
   Vector V;
   ...
   PrintVector(V,"Vect");
********************************************/
void PrintVector(Vector U,char *VarName,unsigned int n){
  PrintVectorFormat(U,VarName,n,"%.6f");
}

void PrintVectorFormat(Vector U,char *VarName,unsigned int n,const char *format){
  unsigned int i;
  char sformat[1024];
  strcpy(sformat,format);
  strcat(sformat," ");

  if (n==0) n=U.dim;
  printf("%s<\033[0;34m%p\033[0m>[%d] :\n",VarName,(void *)U.val,U.dim);
  printf("  ");

  if (2*n>=U.dim)
    for (i=0;i<U.dim;i++)
      printf(sformat,U.val[i]);
  else{
    for (i=0;i<n;i++)
      printf(sformat,U.val[i]);
    printf("\n  ...\n  ");
    for (i=U.dim-n;i<U.dim;i++)
      printf(sformat,U.val[i]);
  }
  printf("\n");
}

/*******************************************
 PrintVector2 :
   Affiche un vecteur
 USAGE :
   Vector V;
   ...
   PrintVector2(V,"Vect");
********************************************/
void PrintVector2(Vector U,char *VarName){
  unsigned int i;

  for (i=0;i<U.dim;i++)
    printf("  %s[%u]=%lf\n",VarName,i,U.val[i]);
}

/*******************************************
 SaveVector :
   Sauve un vecteur dans un fichier
 USAGE :
   Vector V;
   ...
   SaveVector(V,"fic1.dat");
********************************************/
void SaveVector(Vector U,char *FileName){
  unsigned int i;
  FILE *fid;
  fid=fopen(FileName,"w");
  if ( fid == NULL){
    fprintf(stderr,"Ouverture fichier %s impossible dans le fichier :%s ligne : %d",FileName,__FILE__,__LINE__);
    exit(EXIT_FAILURE);
  }
  for (i=0;i<U.dim;i++)
    fprintf(fid,"%.16lf\n",U.val[i]);
  fclose(fid);
}

/*******************************************
 Save2Vector :
   Sauve deux vecteurs de meme dimension
   dans un fichier
 USAGE :
   Vector U,V;
   ...
   Save2Vector(U,V,"fic1.dat");
********************************************/
void Save2Vector(Vector U,Vector V,char *FileName){
  unsigned int i;
  FILE *fid;
  MESSAGEFAILURE( U.dim != V.dim, "dimension failed");
//   if ( U.dim !=V.dim ){
//     fprintf(stderr,"Probleme de dimension %d <> %d :%s ligne : %d",U.dim,V.dim,__FILE__,__LINE__);
//     exit(EXIT_FAILURE);
//   }
  fid=fopen(FileName,"w");
  if ( fid == NULL){
    fprintf(stderr,"Ouverture fichier %s impossible dans le fichier :%s ligne : %d",FileName,__FILE__,__LINE__);
    exit(EXIT_FAILURE);
  }
  for (i=0;i<U.dim;i++)
    fprintf(fid,"%.16lf %.16lf\n",U.val[i],V.val[i]);
  fclose(fid);
}

/*******************************************
 ScalarProd_out :
   Produit scalaire de 2 vecteurs
 USAGE :
   Vector V,W;
   double ps;
   ...
   ps=ScalarProd_out(V,W);
********************************************/
double ScalarProd_out(Vector U, Vector V){
  unsigned int i;
  double s=0.;
  assert(U.dim==V.dim);
  for (i=0;i<U.dim;i++)
    s+=U.val[i]*V.val[i];
  return s;
}

/*******************************************
 ScalarProd :
   Produit scalaire de 2 vecteurs
 USAGE :
   Vector V,W;
   double ps;
   ...
   ScalarProd(V,W,&ps);
********************************************/
void ScalarProd(Vector U, Vector V, double *ps){
  unsigned int i;
  assert(U.dim==V.dim);
  *ps=0.;
  for (i=0;i<U.dim;i++)
    (*ps)+=U.val[i]*V.val[i];
}

/*******************************************
 aUpbV_out :
   Retourne le vecteur a*U+b*V
   ou a,b sont 2 reels et U,V deux vecteurs.
   Le vecteur retourne est alloue dans cette fonction
 USAGE :
   Vector U,V,W;
   ...
   W=aUpbV_out(-1,U,1.,V);
 REMARQUE :
   Que se passe-t'il pour
     V=aUpbV_out(-1,U,1.,V);

   see results of: valgrind –leak-check=full ./a.out
********************************************/
Vector aUpbV_out(double a,Vector U, double b, Vector V){
  unsigned int i;
  Vector W;
  MESSAGEFAILURE( U.dim != V.dim, "dimension failed");
  AllocVector(U.dim,&W);
  for (i=0;i<U.dim;i++)
    W.val[i]=a*U.val[i]+b*V.val[i];
  return W;
}

/*******************************************
 aUpbV :
   Calcul le vecteur W <-a*U+b*V
   ou a,b sont 2 reels et U,V deux vecteurs.
   Le vecteur W est alloue dans cette fonction si besoin
 USAGE :
   Vector U,V,W;
   ...
   aUpbV(-1,U,1.,V,&W);
 REMARQUE :
   Que se passe-t'il pour
     aUpbV(-1,U,1.,V,&V);

   see results of: valgrind
********************************************/
void aUpbV(double a,Vector U, double b, Vector V,Vector *pW){
  unsigned int i;
  MESSAGEFAILURE( U.dim != V.dim, "dimension failed");
  AllocVector(U.dim,pW);
  for (i=0;i<U.dim;i++)
    pW->val[i]=a*U.val[i]+b*V.val[i];
}

/*******************************************
 aUpbV_in :
   Calcul le vecteur W=a*U+b*V
   ou a,b sont 2 reels et U,V deux vecteurs.
   Le vecteur W est déjà alloué.
 USAGE :
   Vector U,V,W;
   ...
   aUpbV_in(-1,U,1.,V,W);
********************************************/
void aUpbV_in(double a,Vector U, double b, Vector V,Vector W){
  unsigned int i;
  MESSAGEFAILURE( U.dim != V.dim, "dimension failed");
  MESSAGEFAILURE( U.dim != W.dim, "dimension failed");
  for (i=0;i<U.dim;i++)
    W.val[i]=a*U.val[i]+b*V.val[i];
}


/*******************************************
 ScaleVector :
   Transforme le vecteur U en a*U
     U <- a*U
   ou a est un reel.
   Le vecteur U est deja alloue
 USAGE :
   Vector U;
   ...
   ScaleVector(-1.,U);
********************************************/
void ScaleVector(double a,Vector U){
  unsigned int i;

  for (i=0;i<U.dim;i++)
    U.val[i]*=a;
}

/*******************************************
 NormVector :
   Calcul la norme p d'un vecteur
 USAGE :
   Vector U;
   double norm;
   ...
   norm=NormVector(V,2);
********************************************/
double NormVector(Vector U, unsigned int p){
  unsigned int i;
  double s=0.;
  for (i=0;i<U.dim;i++)
    s+=pow(fabs(U.val[i]),p);
  return pow(s,1./p);
}

/*******************************************
 NormInfVector :
   Calcul la norme Inf d'un vecteur
 USAGE :
   Vector U;
   double norm;
   ...
   norm=NormInfVector(V);
********************************************/
double NormInfVector(Vector U){
  unsigned int i;
  double max=0.;
  for (i=0;i<U.dim;i++)
    max=MAX(fabs(U.val[i]),max);
  return max;
}

/*******************************************
 fevalVector_out :
   return a Vector V such that V[i]=f(U[i])
 USAGE :
   Vector U,V;
   ...
   V=fevalVector_out(U,cos);
 REMARQUE:
   fuite mémoire possible!
********************************************/
Vector fevalVector_out(Vector U,double (*pf)(double)){
  Vector V;
  unsigned int i;
  AllocVector(U.dim,&V);
  for (i=0;i<U.dim;i++) V.val[i]=pf(U.val[i]);
  return V;
}

/*******************************************
 fevalVector :
   Alloue (si besoin) un Vector V tel que
       V[i]=f(U[i])
 USAGE :
   Vector U,V;
   ...
   fevalVector(U,sin,&V);
 REMARQUE:
   fuite mémoire possible!
********************************************/
void fevalVector(Vector U,double (*pf)(double),Vector *pW){
  unsigned int i;
  AllocVector(U.dim,pW);
  for (i=0;i<U.dim;i++) pW->val[i]=pf(U.val[i]);
}

/*******************************************
 fevalVector_in :
   Modifie le Vecteur U
       U[i] <- f(U[i])
 USAGE :
   Vector U;

   ...
   fevalVector_in(U,exp);
********************************************/
void fevalVector_in(Vector U,double (*pf)(double)){
  unsigned int i;
  for (i=0;i<U.dim;i++) U.val[i]=pf(U.val[i]);
}

/*******************************************
 copyVector :
   Copie intégrale d'un vecteur avec duplication
   du tableau.
 USAGE :
   Vector U,V;

   ...
   copyVector(U,&V);
 REMARQUE:
   fuite mémoire possible!
********************************************/
void copyVector(Vector U,Vector *pW){
  AllocVector(U.dim,pW);
  memcpy(pW->val,U.val,U.dim*sizeof(double));
}

/*******************************************
 copyVector_out :
   Copie intégrale d'un vecteur avec duplication
   du tableau.
 USAGE :
   Vector U,V;
   ...
   V=copyVector_out(U);
 REMARQUE:
   fuite mémoire possible!
********************************************/
Vector copyVector_out(Vector U){
  Vector V;
  AllocVector(U.dim,&V);
  memcpy(V.val,U.val,U.dim*sizeof(double));
  return V;
}
