#ifndef _VECTOR_H_
#define _VECTOR_H_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "slacgen.h"
#include "rand.h"

typedef struct {
 unsigned int dim;
 double *val;
 } Vector;
 
#define V0 {0,NULL}
 
/*******************************************
 AllocVector_out :
   Retourne un vecteur de dimension N 
   avec initialisation a 0 des composantes 
 USAGE :                                 
   Vector V;
   ...
   V=AllocVector_out(10);
  REMARQUE :
    Que se passe-t'il pour:
    Vector V;
    ...
    V=AllocVector_out10);
    V=AllocVector_out(20);
    
    fuite mémoire:  utiliser valgrind
********************************************/
Vector AllocVector_out(unsigned int N);

/*******************************************
 AllocVector :
   Allocation d'un vecteur de dimension N 
   avec initialisation a 0 des composantes 
 USAGE :                                 
   Vector V;
   AllocVector(10,&V);
   
 REMARQUE :
    Que se passe-t'il pour:
    Vector V;
    ...
    AllocVector1(10,&V);
    AllocVector1(20,&V);
    
    fuite mémoire:  utiliser valgrind
********************************************/
void AllocVector(unsigned int N, Vector *pU);

/*******************************************
 secureAllocVector :
   Allocation d'un vecteur de dimension N 
   avec initialisation a 0 des composantes.
   Si le vecteur a deja ete alloue et est 
   de dimension N alors juste une simple mise 
   à 0 du vvecteur.
 USAGE :                                 
   Vector V={0,NULL};
   ...
   secureAllocVector(10,&V);
  REMARQUE :
    n'est pas si secure que cela!
    Vector V;
    secureAllocVector(10,&V);
    
    Core dump!
********************************************/
void secureAllocVector(unsigned int N, Vector *pU);

/*******************************************
 FREEVector :
   Desalloue proprement un vecteur : 
   <pval> <- NULL
   <dim>  <- 0
 USAGE :                                 
   Vector V;
   FREEVector(&V);                      
********************************************/
void FREEVector(Vector *pV);

/*******************************************
 FREEVector :
   Desalloue simplement un vecteur  
 USAGE :                                 
   Vector V;
   freeVector(V);                      
********************************************/
void freeVector(Vector V);

/*******************************************
 Vector2zero :
   Met les composantes d'un vecteur à 0
 USAGE :                                 
   Vector V;
   ...
   Vector2zero(V);                      
********************************************/
void Vector2zero(Vector V);

/*******************************************
 Vector2cst :
   Met les composantes d'un vecteur à c
 USAGE :                                 
   Vector V;
   ...
   Vector2cst(V,1.);                      
********************************************/
void Vector2cst(Vector V,double c);

/*******************************************
 RandVector :
   Initialise les composantes d'un vecteur
   suivant la loi uniforme sur [a;b].
   Le vecteur est suppose deja alloue.
 USAGE :                                 
   Vector V;
   ...
   RandVector(V,-1.,1.);                      
********************************************/
void RandVector(Vector U, double a, double b);

/*******************************************
 AllocRandVector :
   Alloue (si besoin) et initialise les composantes d'un vecteur
   suivant la loi uniforme sur [a;b].
 USAGE :                                 
   Vector V;
   ...
   AllocRandVector(N,-1.,1.,&V);                      
********************************************/
void AllocRandVector(int N, double a, double b,Vector *pW);

/*******************************************
 AllocRandVector_out :
   Alloue un vecteur de dimension N et
   initialise les composantes d'un vecteur
   suivant la loi uniforme sur [a;b].
 USAGE :                                 
   Vector V;
   V=AllocRandVector_out(10,-1.,1.);                      
 REMARQUE :
   Attention aux fuites mémoire
********************************************/
Vector AllocRandVector_out(unsigned int N, double a, double b);

/*******************************************
 PrintVector :
   Affiche un vecteur
 USAGE :                                 
   Vector V;
   ...
   PrintVector(V,"Vect");                      
********************************************/
void PrintVector(Vector U,char *VarName,unsigned int n);

void PrintVectorFormat(Vector U,char *VarName,unsigned int n,const char *format);

void PrintVector2(Vector U,char *VarName);

/*******************************************
 SaveVector :
   Sauve un vecteur dans un fichier
 USAGE :                                 
   Vector V;
   ...
   SaveVector(V,"fic1.dat");                      
********************************************/
void SaveVector(Vector U,char *FileName);

/*******************************************
 Save2Vector :
   Sauve deux vecteurs de meme dimension
   dans un fichier
 USAGE :                                 
   Vector U,V;
   ...
   Save2Vector(U,V,"fic1.dat");                      
********************************************/
void Save2Vector(Vector U,Vector V,char *FileName);

/*******************************************
 ScalarProd_out :
   Produit scalaire de 2 vecteurs
 USAGE :                                 
   Vector V,W;
   double ps;
   ...
   ps=ScalarProd_out(V,W);                      
********************************************/
double ScalarProd_out(Vector U, Vector W);

/*******************************************
 ScalarPro2 :
   Produit scalaire de 2 vecteurs
 USAGE :                                 
   Vector V,W;
   double ps;
   ...
   ScalarProd(V,W,&ps);                      
********************************************/
void ScalarProd(Vector U, Vector W, double *ps);

/*******************************************
 aUpbV_out :
   Retourne le vecteur a*U+b*V
   ou a,b sont 2 reels et U,V deux vecteurs.
   Le vecteur retourne est alloue dans cette fonction
 USAGE :                                 
   Vector U,V,W;
   ...
   W=aUpbV_out(-1,U,1.,V);                      
 REMARQUE :
   Que se passe-t'il pour
     V=aUpbV_out(-1,U,1.,V); 
     
   see results of: valgrind –leak-check=full ./a.out 
********************************************/
Vector aUpbV_out(double a,Vector U, double b, Vector V);

/*******************************************
 aUpbV :
   Calcul le vecteur W <-a*U+b*V
   ou a,b sont 2 reels et U,V deux vecteurs.
   Le vecteur W est alloue dans cette fonction si besoin
 USAGE :                                 
   Vector U,V,W;
   ...
   aUpbV(-1,U,1.,V,&W);
 REMARQUE :
   Que se passe-t'il pour
     aUpbV(-1,U,1.,V,&V); 
     
   see results of: valgrind 
********************************************/
void aUpbV(double a,Vector U, double b, Vector V,Vector *pW);

/*******************************************
 aUpbV_in :
   Calcul le vecteur W=a*U+b*V
   ou a,b sont 2 reels et U,V deux vecteurs.
   Le vecteur W est déjà alloué.
 USAGE :                                 
   Vector U,V,W;
   ...
   aUpbV_in(-1,U,1.,V,W);                      
********************************************/
void aUpbV_in(double a,Vector U, double b, Vector V,Vector W);

/*******************************************
 ScaleVector :
   Transforme le vecteur U en a*U
     U <- a*U
   ou a est un reel.
   Le vecteur U est deja alloue 
 USAGE :                                 
   Vector U;
   ...
   ScaleVector(-1.,U);                      
********************************************/
void ScaleVector(double a,Vector U);

/*******************************************
 NormVector :
   Calcul la norme p d'un vecteur 
 USAGE :                                 
   Vector U;
   double norm;
   ...
   norm=NormVector(V,2);                      
********************************************/
double NormVector(Vector U, unsigned int p);

/*******************************************
 NormInfVector :
   Calcul la norme Inf d'un vecteur 
 USAGE :                                 
   Vector U;
   double norm;
   ...
   norm=NormInfVector(V);                      
********************************************/
double NormInfVector(Vector U);

/*******************************************
 fevalVector_out :
   return a Vector V such that V[i]=f(U[i])
 USAGE :                                
   Vector U,V;
   ...
   V=fevalVector_out(U,cos);
 REMARQUE:
   fuite mémoire possible!
********************************************/
Vector fevalVector_out(Vector U,double (*pf)(double));

/*******************************************
 fevalVector :
   Alloue (si besoin) un Vector V tel que 
       V[i]=f(U[i])
 USAGE :                                 
   Vector U,V;
   ...
   fevalVector(U,sin,&V);
 REMARQUE:
   fuite mémoire possible!
********************************************/
void fevalVector(Vector U,double (*pf)(double),Vector *pW);

/*******************************************
 fevalVector_in :
   Modifie le Vecteur U  
       U[i] <- f(U[i])
 USAGE :                                 
   Vector U;
   
   ...
   fevalVector_in(U,exp);
********************************************/
void fevalVector_in(Vector U,double (*pf)(double));

/*******************************************
 copyVector :
   Copie intégrale d'un vecteur avec duplication
   du tableau.
 USAGE :                                 
   Vector U,V;
   ...
   copyVector(U,&V);
 REMARQUE:
   fuite mémoire possible!
********************************************/
void copyVector(Vector U,Vector *pW);

/*******************************************
 copyVector_out :
   Copie intégrale d'un vecteur avec duplication
   du tableau.
 USAGE :                                 
   Vector U,V;
   ...
   V=copyVector_out(U);
 REMARQUE:
   fuite mémoire possible!
********************************************/
Vector copyVector_out(Vector U);

#endif
